import importlib
from typing import Any, Dict, List, Type

from .kafka_message import KafkaMessage
from .kafka_message_consumer import KafkaMessageConsumer


class MessageTypeRegistryEntry:
    msg_type_path: str | None = None
    msg_type: type | None = None
    _consumers: List[Type[KafkaMessageConsumer]] = []

    def __init__(self, msg_type: Type[KafkaMessage], consumer_type: Type[KafkaMessageConsumer]):
        self.msg_type_path = f"{msg_type.__module__}.{msg_type.__name__}"
        self.msg_type = self.__load_msg_type(self.msg_type_path)
        self._consumers = [consumer_type]

    def register_consumer(self, consumer: Type[KafkaMessageConsumer]):
        if consumer not in self._consumers:
            self._consumers.append(consumer)

    def get_consumers(self) -> List[Type[KafkaMessageConsumer]]:
        return self._consumers

    def purge(self):
        self.msg_type = None
        self._consumers.clear()

    @staticmethod
    def __load_msg_type(msg_type_path: str) -> type:
        module_path, class_name = msg_type_path.rsplit(".", 1)
        module = importlib.import_module(module_path)  # nosemgrep: non-literal-import - loading only child classes of KafkaMessage
        msg_type: type = getattr(module, class_name)
        if not issubclass(msg_type, KafkaMessage):
            raise TypeError(f"{msg_type} is not a subclass of KafkaMessage")
        return msg_type


class MessageTypeRegistry:
    _registry: Dict[str, MessageTypeRegistryEntry] = {}

    def register_message_handler(self, msg_type: Type[KafkaMessage], consumer: Type[KafkaMessageConsumer]):
        msg_type_name = msg_type.__name__
        if self._registry.get(msg_type_name) is None:
            self._registry[msg_type_name] = MessageTypeRegistryEntry(msg_type, consumer)
        else:
            self._registry[msg_type_name].register_consumer(consumer)
        pass

    def get_message_type_from_name(self, msg_type_name):
        if self._registry.get(msg_type_name) is None:
            return None
        return self._registry[msg_type_name].msg_type

    def get_handlers_for_msg_type_name(self, msg_type_name) -> List[Type[KafkaMessageConsumer]] | None:
        if self._registry.get(msg_type_name) is None:
            return None
        return self._registry[msg_type_name].get_consumers()

    def clear(self) -> None:
        for v in self._registry.values():
            v.purge()
        self._registry.clear()


class TopicMessageTypeRegistry:
    _registry: Dict[str, MessageTypeRegistry] = {}

    def register_handler_for_topic(self, topic, message_cls: Type[KafkaMessage], handler_cls: Type[KafkaMessageConsumer]):
        if self._registry.get(topic) is None:
            self._registry[topic] = MessageTypeRegistry()
        self._registry[topic].register_message_handler(message_cls, handler_cls)

    def get_topics(self) -> List[str]:
        return list(self._registry.keys())

    def get_message_type_from_name(self, topic: str, msg_type_name: str) -> Type[KafkaMessage] | None:
        if self._registry.get(topic) is None:
            return None
        return self._registry[topic].get_message_type_from_name(msg_type_name)

    def get_handlers(self, topic, msg_type) -> List[Type[KafkaMessageConsumer[Any]]] | None:
        if self._registry.get(topic) is None:
            return None
        msg_type_name = msg_type.__name__
        return self._registry[topic].get_handlers_for_msg_type_name(msg_type_name)

    def purge(self):
        for v in self._registry.values():
            v.clear()
        self._registry.clear()


_message_type_registry: TopicMessageTypeRegistry = TopicMessageTypeRegistry()



from typing import Type, get_type_hints

from .exceptions import IncompatibleMessageTypeException
from .kafka_message import KafkaMessage
from .kafka_message_consumer import KafkaMessageConsumer
from .message_type_registry import _message_type_registry


def kafka_message_handler(message_cls: Type[KafkaMessage]):
    """
    Register handler from KafkaMessage[T] inherited class,
    ensuring that method`.handle(self, message_data: T)` uses correct T type.
    """

    def wrapper(handler_cls: Type[KafkaMessageConsumer]):
        if not issubclass(handler_cls, KafkaMessageConsumer):
            raise TypeError("Handler must inherit from KafkaMessageConsumer")

        topic = message_cls.topic
        if not topic:
            raise ValueError(f"{message_cls.__name__} must define static field 'topic'")

        # check type from `handle(self, message_data: T)` T parameter
        handle_method = getattr(handler_cls, "handle", None)
        if handle_method is None:
            raise ValueError(f"{handler_cls.__name__} must define a 'handle' method")

        hints = get_type_hints(handle_method)
        if "message_data" not in hints:
            raise ValueError(f"{handler_cls.__name__}.handle must define argument 'message_data' with type annotation")

        expected_type = message_cls.__annotations__.get("payload")
        actual_type = hints["message_data"]
        if actual_type != expected_type:
            raise IncompatibleMessageTypeException(topic, expected_type, actual_type)

        _message_type_registry.register_handler_for_topic(topic, message_cls, handler_cls)
        return handler_cls

    return wrapper



import logging


class DummyLogger(logging.Logger):
    def log(self, level, msg, *args, **kwargs):  # noqa: Vulture
        pass

    def debug(self, msg, *args, **kwargs):  # noqa: Vulture
        pass

    def warning(self, msg, *args, **kwargs):  # noqa: Vulture
        pass

    def error(self, msg, *args, **kwargs):  # noqa: Vulture
        pass

    def critical(self, msg, *args, **kwargs):  # noqa: Vulture
        pass

    def info(self, msg, *args, **kwargs):  # noqa: Vulture
        pass



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Type

from confluent_kafka import KafkaError


@dataclass
class KafkaMessage(ABC):
    topic: str
    payload: object  # to be precised in the inherited class
    _error: KafkaError | None

    def __init__(self) -> None:
        self._error = None

    @abstractmethod
    def get_partition_key(self) -> str:
        """
        The message key in the produce method is important for determining how messages are
        distributed across partitions in a Kafka topic. By using a key, all messages with the same
        key will go to the same partition, which helps maintain the order of messages that are related.

        :return: str
        """
        pass

    def to_json(self) -> str:
        """Serialize the value payload to JSON."""
        payload_type: Type | None = self.__annotations__.get("payload", None)
        if payload_type in {int, float, str, bool, bytes}:
            return f"{self.payload}"
        else:
            return json.dumps(self.payload.__dict__)

    def error(self) -> KafkaError | object | None:
        return self._error

    def set_error(self, error: KafkaError | None) -> None:
        self._error = error



from logging import Logger
from typing import Any, Callable

from confluent_kafka import Message

from .kafka_client import KafkaClient
from .kafka_consumer_manager import KafkaConsumerManager
from .kafka_message import KafkaMessage
from .kafka_message_consumer import KafkaMessageConsumer


class KafkaManager:

    def __init__(
        self,
        kafka_client: KafkaClient,
        kafka_consumer: KafkaConsumerManager,
        logger: Logger,
    ):
        self.kafka_client = kafka_client
        self.kafka_msg_consumer = kafka_consumer
        self.logger = logger

    def set_consumer_provider(self, consumer_provider: Callable[[Any], KafkaMessageConsumer]):
        self.kafka_msg_consumer.set_consumer_provider(consumer_provider)

    def publish_message(
        self,
        message: KafkaMessage,
        on_complete: Callable[[Message, Any | str | Exception | None], None] = lambda m, e: None,  # noqa: Vulture
    ) -> None:
        return self.kafka_client.send(message, on_complete)

    def start_consumer_loop(self):
        self.logger.info("Starting consumer loop")
        self.kafka_msg_consumer.start()

    def stop_consumer_loop(self):
        self.logger.info("Stopping consumer loop")
        self.kafka_client.stop()
        self.kafka_msg_consumer.stop()



import socket
from logging import Logger

from confluent_kafka import Consumer, Producer

from .kafka_client import KafkaClient
from .kafka_consumer_manager import KafkaConsumerManager
from .kafka_manager import KafkaManager
from .retry_policy import RetryPolicy


def kavari_create(
    bootstrap_servers: str,
    group_id: str,
    delivery_retry_policy: RetryPolicy,
    publishing_retry_policy: RetryPolicy,
    logger: Logger,
    auto_commit: bool = False,
    auto_offset_reset: str = "earliest",
    acks: str = "1",  # Options: "0", "1", "all"
):
    return KafkaManager(
        kafka_client=KafkaClient(
            Producer(
                {
                    "bootstrap.servers": bootstrap_servers,
                    "client.id": socket.gethostname(),
                    "acks": acks,
                    "message.timeout.ms": 30000,  # 30 seconds timeout for message delivery
                    "queue.buffering.max.messages": 100000,  # Maximum number of messages allowed in the producer queue
                }
            ),
            publishing_retry_policy,
            logger,
            dlq_suffix=".dlq",
        ),
        kafka_consumer=KafkaConsumerManager(
            Consumer(
                {
                    "bootstrap.servers": bootstrap_servers,
                    "group.id": group_id,
                    "enable.auto.commit": auto_commit,
                    "auto.offset.reset": auto_offset_reset,
                }
            ),
            relaxed=True,
            retry_policy=publishing_retry_policy,
            logger=logger,
        ),
        logger=logger,
    )



from typing import Any


class MalformedMessageException(Exception):
    @staticmethod
    def missing_message_type_header(topic: str, msg: Any) -> Exception:
        return MalformedMessageException(f"Missing message type header on {topic} for message {msg}")


class MissingMessageTypeException(Exception):
    @staticmethod
    def missing_message_type_header(topic: str, msg: Any) -> Exception:
        return MissingMessageTypeException(f"Missing registered message type on {topic} for message {msg}")


class MissingHandlerException(Exception):
    def __init__(self, topic: str, msg_type_name: str):
        super().__init__(f"Missing handler for {msg_type_name} on topic {topic}")


class UnknownMessageTypeException(Exception):
    def __init__(self, topic: str, msg_type_name: str):
        super().__init__(f"Unknown message type {msg_type_name} on topic {topic}")


class IncompatibleMessageTypeException(Exception):
    def __init__(self, topic: str, expected: Any | None, actual: Any | None):
        super().__init__(f"Incompatible message type on topic {topic}: expected {expected}, got {actual}")



from typing import Generic, TypeVar

from .kafka_message import KafkaMessage

T = TypeVar("T", bound=KafkaMessage)


class KafkaMessageConsumer(Generic[T]):
    def handle(self, message_data: T) -> None:  # noqa: Vulture
        raise NotImplementedError



import json
from typing import Any, Type

from dacite import from_dict

from .kafka_message import KafkaMessage


class KafkaMessageDeserializer:
    def deserialize(self, raw_value: bytes, message_cls: Type[KafkaMessage]) -> Any | None:
        payload_type: Type | None = message_cls.__annotations__.get("payload", None)  # Get expected payload type

        # Check if the payload type is a primitive type
        if payload_type in {int, float, bool, bytes}:
            return payload_type(raw_value)
        if payload_type is str:
            return raw_value.decode("utf-8")  # strings

        # Otherwise, deserialize as JSON
        payload_dict = json.loads(raw_value.decode("utf-8"))
        return from_dict(data_class=payload_type, data=payload_dict)  # type: ignore[arg-type]



import threading
import time
from logging import Logger
from typing import Any, Callable, List, Optional, Type

from confluent_kafka import Consumer, KafkaError, Message

from .exceptions import MalformedMessageException, MissingHandlerException, UnknownMessageTypeException
from .kafka_message_consumer import KafkaMessageConsumer
from .kafka_message_deserializer import KafkaMessageDeserializer
from .message_type_registry import _message_type_registry
from .retry_policy import RetryPolicy


class KafkaConsumerManager:
    def __init__(self, consumer: Consumer, retry_policy: RetryPolicy, logger: Logger, relaxed=True):
        self.logger: Logger = logger
        self.relaxed: bool = relaxed
        self.retry_policy: RetryPolicy = retry_policy
        self.consumer_provider: Callable[[Any], KafkaMessageConsumer] | None = None
        self.threadHandle: threading.Thread | None = None
        self.consumer: Consumer = consumer
        self.topics: List[str] = _message_type_registry.get_topics()
        self.stop_event = threading.Event()
        self.consumer.subscribe(self.topics)
        self.message_deserializer = KafkaMessageDeserializer()

    def get_header_value(self, headers: List[tuple], key: str) -> Optional[str]:
        """Retrieve the value of a header by key, or None if not found."""
        header_dict = dict(headers)
        value = header_dict.get(key)
        return value if value is not None else None

    def set_consumer_provider(self, consumer_provider: Callable[[Any], KafkaMessageConsumer]):
        self.consumer_provider = consumer_provider

    def start(self):
        self.consumer.subscribe(topics=self.topics)
        print(f"[kavari] Subscribed to kafka topics: {self.topics}")

        def run():
            while not self.stop_event.is_set():
                msg: Message = self.consumer.poll(1.0)
                self.logger.debug(f"[kavari] Poll result: {msg}")
                # sleep if the queue is empty to not overkill performance with nasty query rate
                if msg is None:
                    time.sleep(1)
                    continue
                if msg.error():
                    if msg.error().code() != KafkaError._PARTITION_EOF:
                        self.logger.error(f"Kafka Error: {msg.error()}")
                    continue

                topic = msg.topic()
                headers = msg.headers()
                msg_type_header = self.get_header_value(headers, "msg_type")
                if not msg_type_header:
                    if self.relaxed:
                        self.logger.warning(f"Missing message type for topic {topic} in message {msg}")
                        self.consumer.commit(msg)
                        continue
                    raise MalformedMessageException.missing_message_type_header(topic, msg)

                msg_type = _message_type_registry.get_message_type_from_name(topic, msg_type_header)
                if msg_type is None:
                    if self.relaxed:
                        self.logger.warning(f"Missing registered message type for topic {topic} in message {msg}")
                        self.consumer.commit(msg)
                        continue
                    raise UnknownMessageTypeException(topic, msg_type_header)

                handler_cls_list: List[Type[KafkaMessageConsumer]] | None = _message_type_registry.get_handlers(topic, msg_type)

                if not handler_cls_list or handler_cls_list is None:
                    self.logger.warning(f"Kafka: No handler registered for topic: {topic}")
                    continue

                handlers_count = len(handler_cls_list)
                if handlers_count == 0:
                    if self.relaxed:
                        self.logger.info(f"Kafka: No handler registered for topic: {topic}. Proceeding to the next message")
                        self.consumer.commit(msg)
                        continue
                    raise MissingHandlerException(topic, msg_type_header)

                for handler_cls in handler_cls_list:
                    if handler_cls is None:
                        self.logger.warning(f"Kafka: Empty handler registered for topic: {topic}")
                        continue

                    try:
                        message_payload: Any = self.message_deserializer.deserialize(msg.value(), msg_type)
                    except Exception as e:
                        self.logger.error(f"Kafka: Unable to deserialize message: {msg.value()}")
                        raise e

                    try:
                        # Resolve handler from Dependency Injection container
                        assert self.consumer_provider is not None  # nosec B101
                        handler = self.consumer_provider(handler_cls)
                        handler.handle(message_payload)

                        # ACK after success
                        self.consumer.commit(msg)
                    except Exception as e:
                        self.logger.error(f"Kafka: Failed to process message on topic {topic}: {e}")
                        self.logger.exception("Exception details")  # Log the full exception details

        self.stop_event.clear()
        self.threadHandle = threading.Thread(target=run, daemon=True, name="Kafka consumer thread")
        self.threadHandle.start()

    def stop(self) -> None:
        self.stop_event.set()
        self.consumer.unsubscribe()
        assert self.threadHandle is not None  # nosec B101
        self.threadHandle.join()
        self.logger.info("Kafka: Consumer thread stopped.")



import time
from logging import Logger
from typing import Callable, Any

from confluent_kafka import Message, Producer

from .kafka_message import KafkaMessage
from .retry_policy import RetryPolicy


class KafkaClient:
    def __init__(self, producer: Producer, retry_policy: RetryPolicy, logger: Logger, dlq_suffix: str = ".dlq"):
        self.logger: Logger = logger
        self.dlq_suffix: str = dlq_suffix
        self.producer: Producer = producer
        self.retry_policy: RetryPolicy = retry_policy

    def send(
        self,
        message: KafkaMessage,
        on_complete: Callable[[Message, Any | str | Exception | None], None],
    ):
        value = message.to_json()
        topic = message.topic
        msg_type = message.__class__.__name__
        retry_delays = self.retry_policy.delays()
        delivery_success = False

        def delivery_callback(err, msg):
            nonlocal delivery_success
            if err:
                self.logger.error(f"[kavari] Producer error: {err}")
            else:
                self.logger.info(f"[kavari] Successfully delivered message to {topic}")
                delivery_success = True
                on_complete(msg, None)

        self.logger.info(f"[kavari] Attempting to produce message to topic: {topic}")
        self.logger.debug(f"[kavari] Message value: {value}")

        original_error: str | Exception | None = None
        for retry_delay in retry_delays:
            try:
                self.producer.produce(
                    topic=topic,
                    value=value,
                    key=message.get_partition_key(),
                    on_delivery=delivery_callback,
                    headers={"msg_type": msg_type},
                )

                # Poll in a loop with a shorter timeout to check delivery status
                poll_start = time.time()
                while time.time() - poll_start < 30:  # 30 second total timeout
                    self.producer.poll(timeout=1.0)
                    if delivery_success:
                        self.logger.info("[kavari] Message delivered successfully")
                        return
                    time.sleep(0.1)

                if delivery_success:
                    self.logger.info("[kavari] Successfully flushed all messages")
                    return
                else:
                    original_error = "[kavari] Message delivery timeout"
                    self.logger.error(original_error)
                    time.sleep(retry_delay)
                    continue  # retry attempt
            except Exception as e:
                self.logger.warning(f"Unable to send message {message} because of {e}")
                original_error = e
                time.sleep(retry_delay)
                continue

        self.logger.error(f"[kavari] All retry attempts failed for topic {topic}. Moving to DLQ.")
        dlq_topic = f"{message.topic}{self.dlq_suffix}"
        self.producer.produce(
            topic=dlq_topic,
            value=value,
            key=message.get_partition_key(),
            headers={"msg_type": msg_type, "original_error": str(original_error)},
            on_delivery=lambda err, msg: on_complete(msg, err or original_error),
        )
        self.producer.flush()

    def stop(self):
        self.logger.info("[kavari] Flushing produced messages, with timeout 30s")
        self.producer.flush(timeout=30)



from .factory import kavari_create
from .fibonacci_retry_policy import FibonacciRetryPolicy
from .kafka_client import KafkaClient
from .kafka_consumer_manager import KafkaConsumerManager
from .kafka_manager import KafkaManager
from .kafka_message import KafkaMessage
from .kafka_message_consumer import KafkaMessageConsumer
from .kafka_message_handler import kafka_message_handler
from .retry_policy import RetryPolicy
from .tooling import DummyLogger

__all__ = [
    "KafkaMessage",
    "KafkaMessageConsumer",
    "KafkaManager",
    "kafka_message_handler",
    "RetryPolicy",
    "FibonacciRetryPolicy",
    "kavari_create",
]



from .retry_policy import RetryPolicy


class FibonacciRetryPolicy(RetryPolicy):
    def __init__(self, max_attempts: int, dlq_suffix: str = ".dlq"):
        self.max_attempts = max_attempts
        self.dlq_suffix = dlq_suffix

    def delays(self) -> list[int]:
        delays = [1, 2]
        while len(delays) < self.max_attempts:
            delays.append(delays[-1] + delays[-2])
        return delays[: self.max_attempts]
